/*
SAIL SIMULATOR
(c) Kif Scheuer 2024
This is a simulation of sailing dynamics to help illustrate 
how to read wind and set sail position using a tiller.
3 motors involved;
1) for fan to generate wind
2) for position of wind
3) for orientation of boat

A potentiometer is used to drive motor position and create tiller. 
There's some drift to tiller when centering to reflect delay in boat motion
The speed of boat turn is proportional to depth of tiller position
The tiller can turn boat 360 degrees
The position for wind is to create options or changing conditions.
bicolor matrix to show mode

Current version uses state machines to allow more dynamic multitasking of tiller and wind motion.
two buttons control wind left and right orientation. 
one buton switches between modes for different features.
1) just tiller
2) tiller and wind adjustment
3) random wind movement
4) swap tiller control from boat to wind
*/


//======for the motors===
#include <Adafruit_MotorShield.h>
Adafruit_MotorShield AFMS = Adafruit_MotorShield();
Adafruit_StepperMotor *myMotorWind = AFMS.getStepper(200, 1);  //stepper motor for wind direction
Adafruit_StepperMotor *myMotorBoat = AFMS.getStepper(200, 2);  //stepper motor for boat position

const int windPin = 10;  //the pin for the fan

int windDirection = FORWARD;  // variable for wind direction
int randomWind;               //variable to set wind randomly during mode 3

int tillerDirection = FORWARD;    // variable for tiller direction controlling boat
int tillerDirection2 = BACKWARD;  // variable for tiller direction controlling windarm
int tillerSpeed = 0;              //variable for tiller speed
int tillerStep = 0;               // variable for steps to do

//======for the potentiometer===
int potpin = 0;               // analog pin used to connect the potentiometer as tiller
int lastPotValue = 0;         // last position of potentiometer
int val;                      // variable to read the value from the analog pin
int valAngle;                 // variable to hold the converted potentiometer reading as an angle
int valSpeed;                 // varialbe to set speed of the boat turning relative to degree of potentiometer angle
int lowAngle = 80;            // this sets the lower end of the window for a "neutral" tiller.
int highAngle = 100;          // this sets the upper end of the window for a "neutral" tiller.
int momentum = 0;             // initial momentum for drift effect.
const int maxMomentum = 18;   // Maximum momentum, controls the drift effect
const int momentumDecay = 1;  // Rate at which momentum decays

//======for the buttons===
const byte buttonPinP = 3;      // the pin that the port pushbutton is attached to
const byte buttonPinS = 2;      // the pin that the starboard pushbutton is attached to
const byte buttonPinM = 4;      // the pin that the wind mode pushbutton is attached to
int buttonStateP = LOW;         // current state of the Port button
int buttonStateS = LOW;         // current state of the Starboard button
int buttonStateM = LOW;         // current state of the Mode button
int buttonStepCounterMax = 50;  // max total steps in any one direction
int buttonModeCounterMax = 5;   // max total modes
int buttonStepCounterP = 0;     // counter to track port steps
int buttonStepCounterS = 0;     // counter to track startboard steps
int buttonModeCounter = 0;      // counter to track mode
int buttonModeLast;             // last state of button for modes.
int buttonStepCounterM1 = 0;    // counter to track the first part of the random movement in mode 3
int buttonStepCounterM2 = 0;    // counter to track the second part of the random movement in mode 3
int buttonStepCounterM3 = 0;    // counter to track the third part of the random movement in mode 3
int buttonStepCounterM4 = 0;    // counter to track the fourth part of the random movement in mode 3
int pauseDuration = 2000;       //For mode three this controls the length of time the wind pauses in one position.
int random1 = 10;               //For mode three this is the lower end of random steps in one direction.
int random2 = 60;               //For mode three this is the upper end of random steps in one direction.


//======for the LED===
#include <Wire.h>
#include <Adafruit_GFX.h>
#include "Adafruit_LEDBackpack.h"

Adafruit_BicolorMatrix matrix = Adafruit_BicolorMatrix();

void setup() {
  if (!AFMS.begin()) {  // create with the default frequency 1.6KHz if (!AFMS.begin(1000)) {  // OR with a different frequency, say 1KHz
    Serial.println("Could not find Motor Shield. Check wiring.");
    while (1)
      ;
  }
  myMotorWind->setSpeed(5);
  myMotorBoat->setSpeed(10);
  pinMode(windPin, OUTPUT);

  pinMode(buttonPinP, INPUT_PULLUP);
  digitalWrite(buttonPinP, HIGH);
  pinMode(buttonPinS, INPUT_PULLUP);
  digitalWrite(buttonPinS, HIGH);
  pinMode(buttonPinM, INPUT_PULLUP);
  digitalWrite(buttonPinM, HIGH);

  randomWind = random(random1, random2);

  matrix.begin(0x74);  // pass in the address

  Serial.begin(9600);
}
static const uint8_t PROGMEM  //bitmaps for the numbers on the display
  m1_bmp[] = { B00011000, B00111000, B01111000, B00011000, B00011000, B00011000, B00011000, B01111110 },
  m2_bmp[] = { B00111000, B01111100, B01000110, B00001100, B00011000, B00110000, B01100010, B01111110 },
  m3_bmp[] = { B00111100, B01100110, B00000110, B00011100, B00011100, B00000110, B01100110, B00111100 },
  m4_bmp[] = { B00001100, B00111100, B00101100, B01001100, B01111110, B00001100, B00001100, B00111110 },
  m5_bmp[] = { B00000000, B00000000, B00000000, B00111100, B00111100, B00000000, B00000000, B00000000 };

void loop() {

  generateWind();

  buttonCheck();
  moveWind();

  setTillerdirection();
  moveTiller();

  ledOutput();
  //textOutput(); //text output is for testing various variable states. Comment out when not using.
}

//======state machines=============

void generateWind() {
  digitalWrite(windPin, HIGH);
}

void buttonCheck() {
  buttonStateP = digitalRead(buttonPinP);
  buttonStateS = digitalRead(buttonPinS);
  buttonStateM = digitalRead(buttonPinM);
  if (buttonStateM != buttonModeLast) {
    if (buttonStateM == HIGH) {
      buttonModeCounter++;
      matrix.clear();
    } else {
      Serial.print("");
    }
  }
  buttonModeLast = buttonStateM;
  if (buttonModeCounter >= buttonModeCounterMax) {
    buttonModeCounter = 0;
  }
  if (buttonModeCounter == 4) {
    buttonStepCounterM1 = 0;
    buttonStepCounterM2 = 0;
    buttonStepCounterM3 = 0;
    buttonStepCounterM4 = 0;
    randomWind = random(random1, random2);
  }
}
void moveWind() {
  //tiller only
  if (buttonModeCounter == 1) {
    myMotorWind->release();
  }

  // tiller with selective wind direction
  if (buttonModeCounter == 2) {
    myMotorWind->setSpeed(5);
    if (buttonStateS == HIGH && (buttonStateP == HIGH)) {
      myMotorWind->release();
    }
    if (buttonStateP == LOW && (buttonStepCounterP <= buttonStepCounterMax)) {
      windDirection = FORWARD;
      myMotorWind->step(1, windDirection, INTERLEAVE);
      buttonStepCounterP++;
      buttonStepCounterS--;
    }
    if (buttonStateS == LOW && (buttonStepCounterS <= buttonStepCounterMax)) {
      windDirection = BACKWARD;
      myMotorWind->step(1, windDirection, INTERLEAVE);
      buttonStepCounterP--;
      buttonStepCounterS++;
    }
  }

  //Random wind mode
  if (buttonModeCounter == 3) {
    myMotorWind->setSpeed(1);
    if (buttonStepCounterM1 < randomWind) {
      myMotorWind->step(1, windDirection, INTERLEAVE);
      buttonStepCounterM1++;
    }
    if (buttonStepCounterM1 >= randomWind && (buttonStepCounterM2 < pauseDuration)) {
      myMotorWind->step(0, windDirection, INTERLEAVE);
      buttonStepCounterM2++;
    }
    if (buttonStepCounterM2 >= pauseDuration && (buttonStepCounterM3 < randomWind)) {
      myMotorWind->step(1, ((windDirection - 3) * -1), INTERLEAVE);
      buttonStepCounterM3++;
    }
    if (buttonStepCounterM3 >= randomWind && (buttonStepCounterM4 < pauseDuration)) {
      myMotorWind->step(0, ((windDirection - 3) * -1), INTERLEAVE);
      buttonStepCounterM4++;
    }
    if (buttonStepCounterM4 >= 300) {
      windDirection = ((windDirection - 3) * -1);
      delay(20);
      buttonStepCounterM1 = 0;
      buttonStepCounterM2 = 0;
      buttonStepCounterM3 = 0;
      buttonStepCounterM4 = 0;
      randomWind = random(random1, random2);
    }
  }
}

void setTillerdirection() {
  val = analogRead(potpin);
  valAngle = map(val, 0, 1023, 0, 180);
  valSpeed = map(val, 0, 1023, 0, 40);
  if (valAngle <= lowAngle) {
    tillerDirection = FORWARD;
    tillerDirection2 = BACKWARD;
    tillerSpeed = ((valSpeed - 20) * -1);
    tillerStep = 1;
    if (val != lastPotValue) {
      momentum = maxMomentum;
      momentum = constrain(momentum, -maxMomentum, maxMomentum);
      lastPotValue = val;
    }
  } else if (valAngle >= highAngle) {
    tillerDirection = BACKWARD;
    tillerDirection2 = FORWARD;
    tillerSpeed = (valSpeed - 20);
    tillerStep = 1;
    if (val != lastPotValue) {
      momentum = maxMomentum;
      momentum = constrain(momentum, -maxMomentum, maxMomentum);
      lastPotValue = val;
    }
  } else if ((valAngle > lowAngle) && (valAngle < highAngle)) {
    if (momentum > 0) {
      momentum -= momentumDecay;
    } else if (momentum < 0) {
      momentum += momentumDecay;
    }
    tillerSpeed = momentum / 6;
    tillerStep = momentum / 6;
  }
}

void moveTiller() {
  if (buttonModeCounter != 4) {
    myMotorBoat->setSpeed(sq(tillerSpeed));
    myMotorBoat->step(tillerStep, tillerDirection, INTERLEAVE);
  }
  if (buttonModeCounter == 4) {
    myMotorWind->setSpeed(1);
    myMotorWind->step(tillerStep, tillerDirection2, INTERLEAVE);
  }
}


void ledOutput() {
  if (buttonModeCounter == 1) {
    matrix.drawBitmap(0, 0, m1_bmp, 8, 8, LED_YELLOW);
    matrix.writeDisplay();
  }
  if (buttonModeCounter == 2) {
    matrix.drawBitmap(0, 0, m2_bmp, 8, 8, LED_GREEN);
    matrix.writeDisplay();
  }
  if (buttonModeCounter == 3) {
    matrix.drawBitmap(0, 0, m3_bmp, 8, 8, LED_GREEN);
    matrix.writeDisplay();
  }
  if (buttonModeCounter == 4) {
    matrix.drawBitmap(0, 0, m4_bmp, 8, 8, LED_GREEN);
    matrix.writeDisplay();
  }
  if (buttonModeCounter == 5) {
    matrix.drawBitmap(0, 0, m5_bmp, 8, 8, LED_RED);
    matrix.writeDisplay();
  }
}

void textOutput() {
  Serial.print(buttonModeCounter);
  Serial.print(" / ");
  Serial.print(momentum);
  Serial.print(" / ");
  Serial.print(tillerStep);
  Serial.print(" / ");
  Serial.print(tillerSpeed);
  Serial.print(" / ");
  Serial.println(valAngle);

  Serial.print(buttonStateM);
  Serial.print(" / ");
  Serial.print(buttonModeCounter);
  Serial.print(" / ");
  Serial.print(buttonModeLast);
  Serial.print(" / ");
  Serial.print(buttonStateM);
  Serial.print(" / ");
  Serial.print(buttonStateP);
  Serial.print(" / ");
  Serial.print(buttonStateS);
  Serial.print(" / ");
  Serial.print(buttonStepCounterP);
  Serial.print(" / ");
  Serial.print(buttonStepCounterS);
  Serial.print(" / ");
  Serial.print(randomWind);
  Serial.print(" / ");
  Serial.print(buttonStepCounterM1);
  Serial.print(" / ");
  Serial.print(buttonStepCounterM2);
  Serial.print(" / ");
  Serial.print(buttonStepCounterM3);
  Serial.print(" / ");
  Serial.print(buttonStepCounterM4);
  Serial.print(" / ");
  Serial.print(windDirection);
  Serial.print(" / ");

  Serial.println("");
}
